<?php

namespace Drupal\authorization_code_sms\Plugin\CodeSender;

use Drupal\authorization_code\Exceptions\FailedToSendCodeException;
use Drupal\authorization_code\Plugin\CodeSender\CodeSenderBase;
use Drupal\Component\Plugin\DependentPluginInterface;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Utility\Token;
use Drupal\sms\Exception\NoPhoneNumberException;
use Drupal\sms\Exception\SmsPluginReportException;
use Drupal\sms\Message\SmsMessage;
use Drupal\sms\Provider\PhoneNumberProviderInterface;
use Drupal\user\UserInterface;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Sms implementation of the code sender plugin.
 *
 * @CodeSender(
 *   id = "sms",
 *   title = @Translation("SMS")
 * )
 */
class Sms extends CodeSenderBase implements ContainerFactoryPluginInterface, DependentPluginInterface {

  /**
   * A logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  private $logger;

  /**
   * Token parser service.
   *
   * @var \Drupal\Core\Utility\Token
   */
  private $token;

  /**
   * Phone number provider service.
   *
   * @var \Drupal\sms\Provider\PhoneNumberProviderInterface
   */
  private $phoneNumberProvider;

  /**
   * Sms constructor.
   *
   * @param array $configuration
   *   The plugin configuration.
   * @param string $plugin_id
   *   The plugin id.
   * @param array $plugin_definition
   *   The plugin definition.
   * @param \Drupal\Core\Utility\Token $token
   *   Token parser service.
   * @param \Drupal\sms\Provider\PhoneNumberProviderInterface $phone_number_provider
   *   Phone number provider service.
   * @param \Psr\Log\LoggerInterface $logger
   *   A logger.
   */
  public function __construct(array $configuration, string $plugin_id, array $plugin_definition, Token $token, PhoneNumberProviderInterface $phone_number_provider, LoggerInterface $logger) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->phoneNumberProvider = $phone_number_provider;
    $this->token = $token;
    $this->logger = $logger;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('token'),
      $container->get('sms.phone_number'),
      $container->get('logger.channel.authorization_code')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function sendCode(UserInterface $user, string $code) {
    $message = (new SmsMessage())
      ->setMessage($this->buildMessage($user, $code));

    try {
      $this->phoneNumberProvider->sendMessage($user, $message);
    }
    catch (SmsPluginReportException $e) {
      $this->logger->error('Failed to send code.<br> Message: %message<br> Trace: <pre>@trace</pre>', [
        '%message' => $e->getMessage(),
        '@trace' => $e->getTraceAsString(),
      ]);
      throw new FailedToSendCodeException($user, $e);
    }
    catch (NoPhoneNumberException $e) {
      $this->logger->error('No phone number found for @name (%id)', [
        '@name' => $user->getDisplayName(),
        '%id' => $user->id(),
      ]);
      throw new FailedToSendCodeException($user, $e);
    }
  }

  /**
   * Builds the message to be sent to the user.
   *
   * @param \Drupal\user\UserInterface $user
   *   The user.
   * @param string $code
   *   The authorization code.
   *
   * @return string
   *   The message.
   */
  private function buildMessage(UserInterface $user, string $code): string {
    return $this->token->replace($this->messageTemplate(), [
      'user' => $user,
      'authorization_code' => $code,
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public function calculateDependencies() {
    return NestedArray::mergeDeep(parent::calculateDependencies(), [
      'plugin' => ['sms.phone.user.user'],
      'module' => ['authorization_code_sms', 'sms'],
    ]);
  }

}
