<?php

namespace Drupal\Tests\authorization_code\Unit;

use Drupal\authorization_code\CodeGeneratorInterface;
use Drupal\authorization_code\CodeRepository;
use Drupal\authorization_code\CodeSenderInterface;
use Drupal\authorization_code\Entity\LoginProcess;
use Drupal\authorization_code\Exceptions\FailedToSaveCodeException;
use Drupal\authorization_code\Exceptions\FailedToSendCodeException;
use Drupal\authorization_code\PluginManager;
use Drupal\authorization_code\UserIdentifierInterface;
use Drupal\Component\Plugin\Exception\PluginException;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\DependencyInjection\ContainerBuilder;
use Drupal\Tests\UnitTestCase;
use Drupal\user\UserInterface;
use phpmock\phpunit\PHPMock;
use Psr\Log\LoggerInterface;

/**
 * Login process test.
 *
 * @group authorization_code
 */
class LoginProcessTest extends UnitTestCase {

  use PHPMock;

  const TEST_CODE = '0123';

  /**
   * Mock of code repository service.
   *
   * @var \Drupal\authorization_code\CodeRepository|\PHPUnit_Framework_MockObject_MockObject
   */
  private $codeRepository;

  /**
   * Mock of code generator plugin manager service.
   *
   * @var \Drupal\authorization_code\PluginManager|\PHPUnit_Framework_MockObject_MockObject
   */
  private $codeGeneratorManager;

  /**
   * Mock of code sender plugin manager service.
   *
   * @var \Drupal\authorization_code\PluginManager|\PHPUnit_Framework_MockObject_MockObject
   */
  private $codeSenderManager;

  /**
   * Mock of user identifier plugin manager service.
   *
   * @var \Drupal\authorization_code\PluginManager|\PHPUnit_Framework_MockObject_MockObject
   */
  private $userIdentifierManager;

  /**
   * Mock of user identifier plugin.
   *
   * @var \Drupal\authorization_code\UserIdentifierInterface|\PHPUnit_Framework_MockObject_MockObject
   */
  private $userIdentifier;

  /**
   * Mock of code sender plugin.
   *
   * @var \Drupal\authorization_code\CodeSenderInterface|\PHPUnit_Framework_MockObject_MockObject
   */
  private $codeSender;

  /**
   * Mock of code generator plugin.
   *
   * @var \Drupal\authorization_code\CodeGeneratorInterface|\PHPUnit_Framework_MockObject_MockObject
   */
  private $codeGenerator;

  /**
   * Mock of a user.
   *
   * @var \Drupal\user\UserInterface|\PHPUnit_Framework_MockObject_MockObject
   */
  private $testUser;

  /**
   * Mock of the container object.
   *
   * @var \Symfony\Component\DependencyInjection\ContainerInterface
   */
  private $container;

  /**
   * {@inheritdoc}
   */
  protected function setUp() {
    parent::setUp();

    $this->userIdentifierManager = $this->createMock(PluginManager::class);
    $this->codeGeneratorManager = $this->createMock(PluginManager::class);
    $this->codeSenderManager = $this->createMock(PluginManager::class);

    $this->codeRepository = $this->createMock(CodeRepository::class);

    $this->userIdentifier = $this->createMock(UserIdentifierInterface::class);
    $this->codeGenerator = $this->createMock(CodeGeneratorInterface::class);
    $this->codeSender = $this->createMock(CodeSenderInterface::class);

    $this->container = new ContainerBuilder();
    $this->container->set('plugin.manager.user_identifier', $this->userIdentifierManager);
    $this->container->set('plugin.manager.code_generator', $this->codeGeneratorManager);
    $this->container->set('plugin.manager.code_sender', $this->codeSenderManager);
    $this->container->set('authorization_code.code_repository', $this->codeRepository);
    $this->container->set('logger.channel.authorization_code', $this->createMock(LoggerInterface::class));
    \Drupal::setContainer($this->container);
  }

  /**
   * Tests startLoginProcess when user identifier not found.
   *
   * @expectedException \Drupal\authorization_code\Exceptions\InvalidLoginProcessConfiguration
   *
   * @throws \Drupal\authorization_code\Exceptions\UserNotFoundException
   * @throws \Drupal\authorization_code\Exceptions\FailedToSaveCodeException
   * @throws \Drupal\authorization_code\Exceptions\FailedToSendCodeException
   */
  public function testStartLoginProcessWhenUserIdentifierNotFound() {
    $this->userIdentifierManager
      ->method('createInstance')
      ->willThrowException(new PluginNotFoundException('test_user_identifier'));

    $this->createLoginProcess()->startLoginProcess('test_user');
  }

  /**
   * Tests completeLoginProcess when user identifier not found.
   *
   * @expectedException \Drupal\authorization_code\Exceptions\InvalidLoginProcessConfiguration
   *
   * @throws \Drupal\authorization_code\Exceptions\InvalidCodeException
   * @throws \Drupal\authorization_code\Exceptions\UserNotFoundException
   */
  public function testCompleteLoginProcessWhenUserIdentifierNotFound() {
    $this->userIdentifierManager
      ->method('createInstance')
      ->willThrowException(new PluginNotFoundException('test_user_identifier'));
    $this->createLoginProcess()
      ->completeLoginProcess('test_user', 'invalid_code');
  }

  /**
   * Tests startLoginProcess when code generator not found.
   *
   * @expectedException \Drupal\authorization_code\Exceptions\InvalidLoginProcessConfiguration
   *
   * @throws \Drupal\authorization_code\Exceptions\UserNotFoundException
   * @throws \Drupal\authorization_code\Exceptions\FailedToSaveCodeException
   * @throws \Drupal\authorization_code\Exceptions\FailedToSendCodeException
   */
  public function testStartLoginProcessWhenCodeGeneratorNotFound() {
    $this->registerUserIdentifierPlugin();
    $this->codeGeneratorManager->method('createInstance')
      ->willThrowException(new PluginException());

    $this->createLoginProcess()->startLoginProcess('test_user');
  }

  /**
   * Tests startLoginProcess when code sender not found.
   *
   * @expectedException \Drupal\authorization_code\Exceptions\InvalidLoginProcessConfiguration
   *
   * @throws \Drupal\authorization_code\Exceptions\UserNotFoundException
   * @throws \Drupal\authorization_code\Exceptions\FailedToSaveCodeException
   * @throws \Drupal\authorization_code\Exceptions\FailedToSendCodeException
   */
  public function testStartLoginProcessWhenCodeSenderNotFound() {
    $this->registerUserIdentifierPlugin();
    $this->registerCodeGeneratorPlugin();
    $this->codeSenderManager->method('createInstance')
      ->willThrowException(new PluginException());

    $this->createLoginProcess()->startLoginProcess('test_user');
  }

  /**
   * Tests startLoginProcess when user is missing.
   *
   * @expectedException \Drupal\authorization_code\Exceptions\UserNotFoundException
   *
   * @throws \Drupal\authorization_code\Exceptions\FailedToSaveCodeException
   * @throws \Drupal\authorization_code\Exceptions\FailedToSendCodeException
   * @throws \Drupal\authorization_code\Exceptions\InvalidLoginProcessConfiguration
   */
  public function testStartLoginProcessWhenUserMissing() {
    $this->registerUserIdentifierPlugin();
    $this->createLoginProcess()->startLoginProcess('missing_user');
  }

  /**
   * Tests startLoginProcess when code save failed.
   *
   * @expectedException \Drupal\authorization_code\Exceptions\FailedToSaveCodeException
   *
   * @throws \Drupal\authorization_code\Exceptions\UserNotFoundException
   * @throws \Drupal\authorization_code\Exceptions\FailedToSendCodeException
   * @throws \Drupal\authorization_code\Exceptions\InvalidLoginProcessConfiguration
   */
  public function testStartLoginProcessWhenSaveFailed() {
    $this->registerUserIdentifierPlugin();
    $this->registerCodeGeneratorPlugin();
    $this->registerCodeSenderPlugin();
    $this->codeRepository
      ->method('saveCode')
      ->with($this->isInstanceOf(LoginProcess::class), $this->testUser, static::TEST_CODE)
      ->willThrowException(new FailedToSaveCodeException());
    $this->createLoginProcess()->startLoginProcess('test_user');
  }

  /**
   * Tests startLoginProcess when code send failed.
   *
   * @expectedException \Drupal\authorization_code\Exceptions\FailedToSendCodeException
   *
   * @throws \Drupal\authorization_code\Exceptions\UserNotFoundException
   * @throws \Drupal\authorization_code\Exceptions\FailedToSaveCodeException
   * @throws \Drupal\authorization_code\Exceptions\InvalidLoginProcessConfiguration
   */
  public function testStartLoginProcessWhenSendFailed() {
    $this->registerUserIdentifierPlugin();
    $this->registerCodeGeneratorPlugin();
    $this->registerCodeSenderPlugin();
    $this->codeRepository
      ->method('saveCode')
      ->with($this->isInstanceOf(LoginProcess::class), $this->testUser, static::TEST_CODE);
    $this->codeSender
      ->method('sendCode')
      ->with($this->testUser, static::TEST_CODE)
      ->willThrowException(new FailedToSendCodeException($this->testUser));
    $this->createLoginProcess()->startLoginProcess('test_user');
  }

  /**
   * Tests startLoginProcess when all is well.
   *
   * @throws \Drupal\authorization_code\Exceptions\FailedToSaveCodeException
   * @throws \Drupal\authorization_code\Exceptions\FailedToSendCodeException
   * @throws \Drupal\authorization_code\Exceptions\InvalidLoginProcessConfiguration
   * @throws \Drupal\authorization_code\Exceptions\UserNotFoundException
   */
  public function testStartLoginProcess() {
    $this->registerUserIdentifierPlugin();
    $this->registerCodeGeneratorPlugin();
    $this->registerCodeSenderPlugin();
    $this->codeSender->expects($this->once())->method('sendCode');
    $this->codeRepository
      ->expects($this->once())
      ->method('saveCode')
      ->with($this->isInstanceOf(LoginProcess::class), $this->testUser, static::TEST_CODE);
    $this->createLoginProcess()->startLoginProcess('test_user');
  }

  /**
   * Tests completeLoginProcess when code is invalid.
   *
   * @expectedException \Drupal\authorization_code\Exceptions\InvalidCodeException
   *
   * @throws \Drupal\authorization_code\Exceptions\InvalidLoginProcessConfiguration
   * @throws \Drupal\authorization_code\Exceptions\UserNotFoundException
   */
  public function testCompleteLoginProcessWhenInvalidCode() {
    $this->registerUserIdentifierPlugin();
    $this->codeRepository->method('isValidCode')->willReturn(FALSE);
    $this->createLoginProcess()
      ->completeLoginProcess('test_user', 'invalid_code');
  }

  /**
   * Test completeLoginProcess when all is well.
   *
   * @throws \Drupal\authorization_code\Exceptions\InvalidCodeException
   * @throws \Drupal\authorization_code\Exceptions\InvalidLoginProcessConfiguration
   * @throws \Drupal\authorization_code\Exceptions\UserNotFoundException
   */
  public function testCompleteLoginProcess() {
    $this->registerUserIdentifierPlugin();
    $this->codeRepository->method('isValidCode')
      ->with($this->isInstanceOf(LoginProcess::class), $this->testUser, static::TEST_CODE)
      ->willReturn(TRUE);
    $this->getFunctionMock('Drupal\authorization_code\Entity', 'user_login_finalize')
      ->expects($this->once())
      ->with($this->testUser);
    $this->createLoginProcess()
      ->completeLoginProcess('test_user', static::TEST_CODE);
  }

  /**
   * Registers the user identifier plugin with the plugin manager mock.
   */
  private function registerUserIdentifierPlugin() {
    $this->testUser = $this->createMock(UserInterface::class);
    $this->userIdentifier
      ->method('getPluginId')
      ->willReturn('test_user_identifier');
    $this->userIdentifier
      ->method('loadUser')->with($this->isType('string'))
      ->will($this->returnValueMap([['test_user', $this->testUser]]));

    $this->userIdentifierManager->method('createInstance')
      ->with('test_user_identifier', ['plugin_id' => 'test_user_identifier'])
      ->willReturn($this->userIdentifier);
  }

  /**
   * Registers the code generator plugin with the plugin manager mock.
   */
  private function registerCodeGeneratorPlugin() {
    $this->codeGenerator->method('getPluginId')->willReturn('test_code_generator');
    $this->codeGenerator->method('generate')->willReturn(static::TEST_CODE);
    $this->codeGeneratorManager->method('createInstance')
      ->with('test_code_generator', ['plugin_id' => 'test_code_generator'])
      ->willReturn($this->codeGenerator);
  }

  /**
   * Registers the code sender plugin with the plugin manager mock.
   */
  private function registerCodeSenderPlugin() {
    $this->codeSender->method('getPluginId')->willReturn('test_code_sender');
    $this->codeSenderManager->method('createInstance')
      ->with('test_code_sender', [
        'plugin_id' => 'test_code_sender',
        'settings' => ['message_template' => 'test message'],
      ])
      ->willReturn($this->codeSender);
  }

  /**
   * Creates a login process entity.
   *
   * @return \Drupal\authorization_code\Entity\LoginProcess
   *   The login process entity.
   */
  private function createLoginProcess(): LoginProcess {
    return new LoginProcess([
      'id' => 'test_login_process',
      'user_identifier' => ['plugin_id' => 'test_user_identifier'],
      'code_generator' => ['plugin_id' => 'test_code_generator'],
      'code_sender' => [
        'plugin_id' => 'test_code_sender',
        'settings' => [
          'message_template' => 'test message',
        ],
      ],
    ], 'login_process');
  }

}
